import copy
import glob
import inspect
import logging
import os
import tempfile
import uuid
from inspect import Parameter
from typing import Any, Dict, List, Optional, Type, Union, cast

import numpy
import yaml
from allennlp.common import Params
from allennlp.data import Vocabulary
from allennlp.models import load_archive
from allennlp.models.archival import Archive
from biome.text import vocabulary
from biome.text.configuration import (
    PipelineConfiguration,
    TrainerConfiguration,
    VocabularyConfiguration,
)
from biome.text.data import DataSource
from biome.text.errors import ActionNotSupportedError, EmptyVocabError
from biome.text.helpers import update_method_signature, save_dict_as_yaml
from dask import dataframe as dd

from . import constants
from ._configuration import (
    ElasticsearchExplore,
    ExploreConfiguration,
    TrainConfiguration,
    _ModelImpl,
)
from .backbone import ModelBackbone
from .modules.heads import TaskHead, TaskHeadSpec

try:
    import ujson as json
except ModuleNotFoundError:
    import json

logging.getLogger("allennlp").setLevel(logging.ERROR)
logging.getLogger("elasticsearch").setLevel(logging.ERROR)


class Pipeline:
    """Manages NLP models configuration and actions.

    Use `Pipeline` for creating new models from a configuration or loading a pre-trained model.

    Use instantiated Pipelines for training from scratch, fine-tuning, predicting, serving, or exploring predictions.
    """

    __LOGGER = logging.getLogger(__name__)
    __TRAINING_CACHE_DATA = "instances_data"

    _model: _ModelImpl = None
    _config: PipelineConfiguration = None

    @classmethod
    def from_yaml(cls, path: str, vocab_path: Optional[str] = None) -> "Pipeline":
        """Creates a pipeline from a config yaml file path

        Parameters
        ----------
        path: `str`
            The path to a YAML configuration file
        vocab_path: `Optional[str]`
            If provided, the pipeline vocab will be loaded from this path

        Returns
        -------
        pipeline: `Pipeline`
            A configured pipeline
        """
        with open(path) as yaml_file:
            config_dict = yaml.safe_load(yaml_file)

        return cls.from_config(config_dict, vocab_path=vocab_path)

    @classmethod
    def from_config(
        cls,
        config: Union[PipelineConfiguration, dict],
        vocab_path: Optional[str] = None,
    ) -> "Pipeline":
        """Creates a pipeline from a `PipelineConfiguration` object

        Parameters
        ----------
        config: `Union[PipelineConfiguration, dict]`
            A `PipelineConfiguration` object or a configuration dict
        vocab_path: `Optional[str]`
            If provided, the pipeline vocab will be loaded from this path

        Returns
        -------
        pipeline: `Pipeline`
            A configured pipeline
        """

        if isinstance(config, dict):
            config = PipelineConfiguration.from_params(Params(config))
        return _BlankPipeline(
            config=config, vocab=vocabulary.load_vocabulary(vocab_path)
        )

    @classmethod
    def from_pretrained(cls, path: str, **kwargs) -> "Pipeline":
        """Loads a pipeline from a pre-trained pipeline from a model.tar.gz file path

        Parameters
        ----------
            path: `str`
                The path to the model.tar.gz file of a pre-trained `Pipeline`

        Returns
        -------
            pipeline: `Pipeline`
                A configured pipeline
        """
        return _PreTrainedPipeline(pretrained_path=path, **kwargs)

    def init_prediction_logger(self, output_dir: str, max_logging_size: int = 100):
        """Initialize the prediction logging.

        If initialized, all predictions will be logged to a file called *predictions.json* in the `output_dir`.

        Parameters
        ----------
        output_dir: str
            Path to the folder in which we create the *predictions.json* file.
        max_logging_size: int
            Max disk size to use for prediction logs
        """
        max_bytes = max_logging_size * 1000000
        max_bytes_per_file = 2000000
        n_backups = int(max_bytes / max_bytes_per_file)
        self._model.init_prediction_logger(
            output_dir, max_bytes=max_bytes_per_file, backup_count=n_backups
        )

    def init_prediction_cache(self, max_size: int) -> None:
        """Initialize cache for input predictions

        Parameters
        ----------
        max_size
            Save up to max_size most recent (inputs).
        """
        self._model.init_prediction_cache(max_size)

    def train(
        self,
        output: str,
        training: DataSource,
        trainer: Optional[TrainerConfiguration] = None,
        validation: Optional[DataSource] = None,
        test: Optional[DataSource] = None,
        extend_vocab: Optional[VocabularyConfiguration] = None,
        restore: bool = False,
    ) -> None:
        """Launches a training run with the specified configurations and data sources

        Parameters
        ----------
        output: `str`
            The experiment output path
        training: `DataSource`
            The training data source
        trainer: `TrainerConfiguration`
            The trainer file path
        validation: `Optional[DataSource]`
            The validation data source
        test: `Optional[DataSource]`
            The test data source
        extend_vocab: `Optional[VocabularyConfiguration]`
            Extends vocab tokens with provided configuration
        restore: `bool`
            If enabled, tries to read previous training status from output folder and
            continues training process from it
        """
        trainer = trainer or TrainerConfiguration()

        from ._helpers import _allennlp_configuration

        allennlp_logger = logging.getLogger("allennlp")

        try:
            allennlp_logger.setLevel(logging.INFO)

            self.__prepare_experiment_folder(output, restore)
            # The original pipeline keeps unchanged
            model = copy.deepcopy(self._model)
            # creates the output folder if it does not exist
            model.cache_data(os.path.join(output, self.__TRAINING_CACHE_DATA))

            vocab = None
            if restore:
                vocab = vocabulary.load_vocabulary(os.path.join(output, "vocabulary"))
            if extend_vocab and not vocab:
                vocab = self._extend_vocabulary(model.vocab, vocab_config=extend_vocab)
            if vocab:
                model.set_vocab(vocab)

            if vocabulary.is_empty(model.vocab, self.config.features.keys):
                raise EmptyVocabError(
                    "Found an empty vocabulary. "
                    "You probably forgot to create a vocabulary with '.create_vocabulary()'."
                )

            # `_allennlp_configuration` needs strings
            datasources_dir = os.path.join(output, "datasources")
            training = training.to_yaml(
                os.path.join(
                    datasources_dir, f"training_{os.path.basename(training.source)}.yml"
                ),
                make_source_path_absolute=True,
            )
            if validation is not None:
                validation = validation.to_yaml(
                    os.path.join(
                        datasources_dir,
                        f"validation_{os.path.basename(validation.source)}.yml",
                    ),
                    make_source_path_absolute=True,
                )
            if test is not None:
                test = test.to_yaml(
                    os.path.join(
                        datasources_dir, f"test_{os.path.basename(test.source)}.yml"
                    ),
                    make_source_path_absolute=True,
                )

            config = TrainConfiguration(
                output=output,
                trainer=trainer,
                training=training,
                validation=validation,
                test=test,
            )

            model.launch_experiment(
                params=Params(_allennlp_configuration(self, config)),
                serialization_dir=output,
            )
        finally:
            allennlp_logger.setLevel(logging.ERROR)

    def __prepare_experiment_folder(self, output: str, restore: bool) -> None:
        """Prepare experiment folder depending of if required experiment restore or not

        Parameters
        ----------
        output
            Path to the output folder
        restore: `bool`
            If False, drops all previous training states

        """
        if not os.path.isdir(output):
            return

        drop_patterns = [
            os.path.join(output, "*.json"),
            os.path.join(output, "**/events.out*"),
        ]

        if not restore:
            drop_patterns.append(os.path.join(output, "*.th"))
            drop_patterns.append(os.path.join(output, self.__TRAINING_CACHE_DATA, "*"))

        for pattern in drop_patterns:
            for file in glob.glob(pattern, recursive=True):
                os.remove(file)

    def predict(self, *args, **kwargs) -> Dict[str, numpy.ndarray]:
        """Predicts over some input data with current state of the model

        The accepted input is dynamically calculated from head.input

        Returns
        -------
        predictions: `Dict[str, numpy.ndarray]`
            A dictionary containing the predictions and additional information
        """
        return self._model.predict(*args, **kwargs)

    def explain(self, *args, **kwargs) -> Dict[str, Any]:
        """Predicts over some input data with current state of the model and provides explanations of token importance.

        The accepted input is dynamically calculated from head.input

        Returns
        -------
        predictions: `Dict[str, numpy.ndarray]`
            A dictionary containing the predictions with token importance calculated using IntegratedGradients
        """
        return self._model.explain(*args, **kwargs)

    def save_vocabulary(self, path: str) -> None:
        """Save the pipeline vocabulary into a path"""
        self._model.vocab.save_to_files(path)

    def create_vocabulary(self, config: VocabularyConfiguration) -> None:
        """Creates a vocabulary an set it to pipeline"""
        raise NotImplementedError

    def explore(
        self,
        data_source: DataSource,
        explore_id: Optional[str] = None,
        es_host: Optional[str] = None,
        batch_size: int = 500,
        prediction_cache_size: int = 0,
        explain: bool = False,
        force_delete: bool = True,
        **metadata,
    ) -> dd.DataFrame:
        """Launches Explore UI for a given data source with current model

        Running this method inside a an `IPython` notebook will try to render the UI directly in the notebook.

        Running this outside a notebook will try to launch the standalone web application.

        Parameters
        ----------
            data_source: `DataSource`
                The data source or its yaml file path
            explore_id: `Optional[str]`
                A name or id for this explore run, useful for running and keep track of several explorations
            es_host: `Optional[str]`
                The URL to the Elasticsearch host for indexing predictions (default is `localhost:9200`)
            batch_size: `int`
                The batch size for indexing predictions (default is `500)
            prediction_cache_size: `int`
                The size of the cache for caching predictions (default is `0)
            explain: `bool`
                Whether to extract and return explanations of token importance (default is `False`)
            force_delete: `bool`
                Deletes exploration with the same `explore_id` before indexing the new explore items (default is `True)

        Returns
        -------
            pipeline: `Pipeline`
                A configured pipeline
        """
        from ._helpers import (
            _explore,
            _show_explore,
        )

        config = ExploreConfiguration(
            batch_size=batch_size,
            prediction_cache_size=prediction_cache_size,
            explain=explain,
            force_delete=force_delete,
            **metadata,
        )

        es_config = ElasticsearchExplore(
            es_index=explore_id or str(uuid.uuid1()),
            es_host=es_host or constants.DEFAULT_ES_HOST,
        )

        explore_df = _explore(self, data_source, config, es_config)
        _show_explore(es_config)

        return explore_df

    def serve(self, port: int = 9998):
        """Launches a REST prediction service with current model in a specified port (default is `9998)

        # Parameters
            port: `int`
                The port to make available the prediction service
        """
        from ._helpers import _serve

        self._model = self._model.eval()
        return _serve(self, port)

    def set_head(self, type: Type[TaskHead], **params):
        """Sets a new task head for the pipeline

        Use this to reuse the weights and config of a pre-trained model (e.g., language model) for a new task.

        Parameters
        ----------
        type: `Type[TaskHead]`
            The `TaskHead` class to be set for the pipeline (e.g., `TextClassification`
        params: `**kwargs`
            The `TaskHead` specific parameters (e.g., classification head needs a `pooler` layer)
        """

        self._config.head = TaskHeadSpec(type=type.__name__, **params)
        self._model.set_head(self._config.head.compile(backbone=self.backbone))

    @property
    def name(self):
        """Gets pipeline  name"""
        return self._model.name

    @property
    def inputs(self) -> List[str]:
        """Gets pipeline input field names"""
        return self._model.inputs

    @property
    def output(self) -> str:
        """Gets pipeline output field names"""
        return self._model.output

    @property
    def backbone(self) -> ModelBackbone:
        """Gets pipeline backbone encoder"""
        return self.head.backbone

    @property
    def head(self) -> TaskHead:
        """Gets pipeline task head"""
        return self._model.head

    @property
    def config(self) -> PipelineConfiguration:
        return self._config

    @property
    def type_name(self) -> str:
        """The pipeline name. Equivalent to task head name"""
        return self.head.__class__.__name__

    @property
    def trainable_parameters(self) -> int:
        """
        Return the number of trainable parameters.

        This number could be change before an after a training process, since trainer could fix some of them.

        """
        return sum(p.numel() for p in self._model.parameters() if p.requires_grad)

    @property
    def trainable_parameter_names(self) -> List[str]:
        """Returns the name of pipeline trainable parameters"""
        return [name for name, p in self._model.named_parameters() if p.requires_grad]

    def _update_prediction_signatures(self):
        """For interactive work-flows, fixes the predict signature to the model inputs"""
        new_signature = inspect.Signature(
            [
                Parameter(name=_input, kind=Parameter.POSITIONAL_OR_KEYWORD)
                for _input in self.inputs
            ]
        )

        for method in [self.predict, self.explain]:
            self.__setattr__(
                method.__name__, update_method_signature(new_signature, method)
            )

    def _extend_vocabulary(
        self, vocab: Vocabulary, vocab_config: VocabularyConfiguration
    ) -> Vocabulary:
        """
        Extends a data vocabulary from a given configuration

        Parameters
        ----------
        vocab: `Vocabulary`
            The source vocabulary
        vocab_config: `VocabularyConfiguration`
            The vocab extension configuration

        Returns
        -------
        vocab: `Vocabulary`
            An extended `Vocabulary` using the provided configuration

        """
        source_paths = [
            source.to_yaml(
                tempfile.NamedTemporaryFile(delete=False).name,
                make_source_path_absolute=True,
            )
            for source in vocab_config.sources
        ]
        instances_vocab = Vocabulary.from_instances(
            instances=[
                instance
                for data_source in source_paths
                for instance in self._model.read(data_source)
            ],
            max_vocab_size=vocab_config.max_vocab_size,
            min_count=vocab_config.min_count,
            pretrained_files=vocab_config.pretrained_files,
            only_include_pretrained_words=vocab_config.only_include_pretrained_words,
            min_pretrained_embeddings=vocab_config.min_pretrained_embeddings,
            tokens_to_add=vocab_config.tokens_to_add,
        )
        vocab.extend_from_vocab(instances_vocab)
        return vocab

    def to_dict(self) -> dict:
        """Returns the pipeline configuration as a dictionary

        Returns
        -------
        pipeline_configuration : dict
        """
        config_dict = copy.deepcopy(self.config.as_dict())
        config_dict["features"]["word"] = (
            config_dict["features"]["word"].to_dict()
            if config_dict["features"]["word"] is not None
            else None
        )
        config_dict["features"]["char"] = (
            config_dict["features"]["char"].to_dict()
            if config_dict["features"]["char"] is not None
            else None
        )

        return config_dict

    def to_yaml(self, path: str):
        """Saves the pipeline configuration to a yaml formatted file

        Parameters
        ----------
        path : str
            Path to the output file
        """
        save_dict_as_yaml(self.to_dict(), path)


class _BlankPipeline(Pipeline):
    """
    Parameters
    ----------
        config: `Optional[PipelineConfiguration]`
            A `PipelineConfiguration` object defining the configuration of the fresh `Pipeline`.

    """

    def __init__(self, config: PipelineConfiguration, **extra_args):
        self._config = config
        self._model = self.__model_from_config(self._config, **extra_args)
        if not isinstance(self._model, _ModelImpl):
            raise TypeError(f"Cannot load model. Wrong format of {self._model}")
        self._update_prediction_signatures()

    @staticmethod
    def __model_from_config(
        config: PipelineConfiguration, **extra_params
    ) -> _ModelImpl:
        """Creates a internal base model from pipeline configuration"""
        return _ModelImpl.from_params(Params({"config": config}), **extra_params)

    def train(
        self,
        output: str,
        training: DataSource,
        trainer: Optional[TrainerConfiguration] = None,
        validation: Optional[DataSource] = None,
        test: Optional[DataSource] = None,
        extend_vocab: Optional[VocabularyConfiguration] = None,
        restore: bool = False,
    ) -> None:
        if extend_vocab:
            raise ActionNotSupportedError(
                "If you want to customize pipeline vocab, please use create_vocab method instead"
            )
        super(_BlankPipeline, self).train(
            output=output,
            training=training,
            trainer=trainer,
            validation=validation,
            test=test,
            restore=restore,
        )

    def create_vocabulary(self, config: VocabularyConfiguration) -> None:
        vocab = self._extend_vocabulary(Vocabulary(), config)
        # TODO: on Allennlp 1.0 we can use model.set_vocab.
        #  For now, we must reload the model passing vocab to allow restore vocab in train
        # self._model.set_vocab(vocab)
        self._model = self.__model_from_config(self.config, vocab=vocab)


class _PreTrainedPipeline(Pipeline):
    """

    Arguments
    ---------
    pretrained_path: `Optional[str]`
        The path to the model.tar.gz of a pre-trained `Pipeline`

    """

    def __init__(self, pretrained_path: str, **extra_args):
        self._binary = pretrained_path
        archive = load_archive(self._binary, **extra_args)
        self._model = self.__model_from_archive(archive)
        self._config = self.__config_from_archive(archive)

        if not isinstance(self._model, _ModelImpl):
            raise TypeError(f"Cannot load model. Wrong format of {self._model}")
        self._update_prediction_signatures()

    def create_vocabulary(self, config: VocabularyConfiguration) -> None:
        raise ActionNotSupportedError(
            "Cannot create a vocabulary for an already pretrained model!!!"
        )

    @staticmethod
    def __model_from_archive(archive: Archive) -> _ModelImpl:
        if not isinstance(archive.model, _ModelImpl):
            raise ValueError(f"Wrong pipeline model: {archive.model}")
        return cast(_ModelImpl, archive.model)

    @staticmethod
    def __config_from_archive(archive: Archive) -> PipelineConfiguration:
        config = archive.config["model"]["config"]
        return PipelineConfiguration.from_params(config)

    @property
    def trained_path(self) -> str:
        """Path to binary file when load from binary"""
        return self._binary



import logging
import os

logging.basicConfig(level=logging.INFO)
test_logger = logging.getLogger(__name__)

TESTS_BASEPATH = os.path.dirname(__file__)
RESOURCES_PATH = os.path.join(TESTS_BASEPATH, "resources")
from .test_support import *



class InmutableDict(dict):
    """Inmutable data dictionary version"""

    def __hash__(self):
        return id(self)

    def _immutable(self, *args, **kws):
        raise TypeError("object is immutable")

    __setitem__ = _immutable
    __delitem__ = _immutable
    __setattr__ = _immutable

    clear = _immutable
    update = _immutable
    setdefault = _immutable
    pop = _immutable
    popitem = _immutable



from typing import List, Optional, Union

from allennlp.data import Instance

from biome.text.backbone import ModelBackbone
from biome.text.modules.specs import (
    ComponentSpec,
    FeedForwardSpec,
    Seq2SeqEncoderSpec,
    Seq2VecEncoderSpec,
)
from .doc_classification import DocumentClassification


class RecordClassification(DocumentClassification):
    """
    Task head for data record  classification.
    Accepts a variable data inputs and apply featuring over defined record keys.

    This head applies a doc2vec architecture from a structured record data input
    """

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        record_keys: List[str],
        tokens_pooler: Optional[Seq2VecEncoderSpec] = None,
        fields_encoder: Optional[Seq2SeqEncoderSpec] = None,
        fields_pooler: Optional[Seq2VecEncoderSpec] = None,
        feedforward: Optional[Seq2SeqEncoderSpec] = None,
        multilabel: Optional[bool] = False,
    ) -> None:

        super(RecordClassification, self).__init__(
            backbone,
            labels=labels,
            tokens_pooler=tokens_pooler,
            sentences_encoder=fields_encoder,
            sentences_pooler=fields_pooler,
            feedforward=feedforward,
            multilabel=multilabel,
        )

        self._inputs = record_keys

    def inputs(self) -> Optional[List[str]]:
        """The inputs names are determined by configured record keys"""
        return self._inputs

    def featurize(
        self, label: Optional[Union[List[str], List[int], str, int]] = None, **inputs
    ) -> Optional[Instance]:

        instance = self.backbone.featurizer(
            {input_key: inputs[input_key] for input_key in self._inputs},
            to_field="document",
        )
        return self.add_label(instance, label)


class RecordClassificationSpec(ComponentSpec[RecordClassification]):
    """Lazy initialization for document classification head components"""

    pass



import click
from click import Path

from biome.text import Pipeline


@click.command("serve", help="Serves pipeline as rest api service")
@click.argument("pipeline_path", type=Path(exists=True))
@click.option("-p", "--port", "port", type=int, default=8888, show_default=True)
def serve(pipeline_path: str, port: int) -> None:
    pipeline = Pipeline.from_pretrained(pipeline_path)
    pipeline.serve(port)



from .task_head import TaskHead, TaskHeadSpec, TaskName, TaskOutput

from .classification.doc_classification import DocumentClassification
from .classification.record_classification import RecordClassification
from .classification.record_pair_classification import RecordPairClassification
from .classification.text_classification import TextClassification

from .language_modelling import LanguageModelling, LanguageModellingSpec
from .token_classification import TokenClassification, TokenClassificationSpec


for head in [
    TextClassification,
    TokenClassification,
    DocumentClassification,
    RecordClassification,
    LanguageModelling,
    RecordPairClassification,
]:
    head.register(overrides=True)



ENV_DASK_CLUSTER = "DASK_CLUSTER"
ENV_DASK_CACHE_SIZE = "DASK_CACHE_SIZE"
DEFAULT_DASK_CACHE_SIZE = 2e9

ENV_ES_HOSTS = "ES_HOSTS"

ES_HOST = "ES_HOST"


BIOME_EXPLORE_ENDPOINT = "BIOME_EXPLORE_ENDPOINT"

CUDA_DEVICE = "CUDA_DEVICE"



from enum import Enum
from typing import Any, Dict, List, Optional

import numpy
import torch
from allennlp.common import Registrable
from allennlp.data import Instance

from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.modules.specs import ComponentSpec


class TaskOutput:
    """
    Task output data class

    A task output will contains almost the logits and probs properties
    """

    def __init__(
        self, logits: torch.Tensor, loss: Optional[torch.Tensor] = None, **extra_data
    ):
        self.logits = logits
        self.loss = loss

        for k, v in extra_data.items():
            self.__setattr__(k, v)

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def as_dict(self) -> Dict[str, torch.Tensor]:
        """Dict reprentation of task output"""
        return {k: v for k, v in self.__dict__.items() if v is not None}


class TaskName(Enum):
    """The task name enum structure"""

    none = 0
    text_classification = 1
    token_classification = 2
    language_modelling = 3

    def __str__(self) -> str:
        return str(self.name)

    def as_string(self) -> str:
        return str(self)


class TaskHead(torch.nn.Module, Registrable):
    """Base task head class"""

    def __init__(self, backbone: ModelBackbone):
        super(TaskHead, self).__init__()
        self.backbone = backbone

    def on_vocab_update(self):
        """
        Actions when vocab is updated. Rebuild here modules that initialization depends on some vocab metric

        At this point, the model.vocab is already updated, so it could be used for architecture update
        """
        pass

    @classmethod
    def register(cls, overrides: bool = False, **kwargs):
        """Enables the task head component for pipeline loading"""
        super(TaskHead, TaskHead).register(cls.__name__, exist_ok=overrides)(cls)

    @property
    def labels(self) -> List[str]:
        """The configured vocab labels"""
        return vocabulary.get_labels(self.backbone.vocab)

    @property
    def num_labels(self):
        """The number of vocab labels"""
        return len(self.labels)

    def extend_labels(self, labels: List[str]):
        """Extends the number of labels"""
        vocabulary.extend_labels(self.backbone.vocab, labels)

    def task_name(self) -> TaskName:
        """The task head name"""
        raise NotImplementedError

    def inputs(self) -> Optional[List[str]]:
        """
        The expected inputs names for data featuring. If no defined,
        will be automatically calculated from featurize signature
        """
        return None

    def forward(self, *args: Any, **kwargs: Any) -> TaskOutput:
        raise NotImplementedError

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        """Metrics dictionary for training task"""
        raise NotImplementedError

    def featurize(self, *args, **kwargs) -> Optional[Instance]:
        """Converts incoming data into an allennlp `Instance`, used for pyTorch tensors generation"""
        raise NotImplementedError

    def decode(self, output: TaskOutput) -> TaskOutput:
        """Completes the output for the prediction

        The base implementation adds nothing.

        Parameters
        ----------
        output
            The output from the head's forward method

        Returns
        -------
        completed_output
        """
        return output

    def explain_prediction(
        self, prediction: Dict[str, numpy.array], instance: Instance
    ) -> Dict[str, Any]:
        """Adds embedding explanations information to prediction output"""
        raise {**prediction, "explain": {}}


class TaskHeadSpec(ComponentSpec[TaskHead]):
    """Layer spec for TaskHead components"""

    pass



import copy
import re
from typing import Callable, Dict, List

from allennlp.common import Registrable
from bs4 import BeautifulSoup


class TextCleaning(Registrable):
    """Base class for text cleaning processors"""

    default_implementation = "default"

    def __call__(self, text: str) -> str:
        """Apply the text transformation"""
        raise NotImplementedError


class TextCleaningRule:
    """Registers a function as a rule for the default text cleaning implementation
    
    Use the decorator `@TextCleaningRule` for creating custom text cleaning and pre-processing rules.
    
    An example function to strip spaces (already included in the default `TextCleaning` processor):
    
    ```python
    @TextCleaningRule
    def strip_spaces(text: str) -> str:
        return text.strip()
    ```
    
    # Parameters
        func: `Callable[[str]`
            The function to register
    """

    __REGISTERED_RULES = {}

    def __init__(self, func: Callable[[str], str]):
        self.__REGISTERED_RULES[func.__name__] = func

    @classmethod
    def registered_rules(cls) -> Dict[str, Callable[[str], str]]:
        """Registered rules dictionary"""
        return copy.deepcopy(cls.__REGISTERED_RULES)


@TextCleaning.register(TextCleaning.default_implementation)
class DefaultTextCleaning(TextCleaning):
    """Defines rules that can be applied to the text before it gets tokenized.

    Each rule is a simple python function that receives and returns a `str`.
    
    # Parameters
        rules: `List[str]`
            A list of registered rule method names to be applied to text inputs
    """

    def __init__(self, rules: List[str] = None):
        self.rules = rules or []
        for rule in self.rules:
            if rule not in TextCleaningRule.registered_rules():
                raise AttributeError(
                    f"No rule '{rule}' registered"
                    f"Available rules are [{[k for k in TextCleaningRule.registered_rules().keys()]}]"
                )

    def __call__(self, text: str) -> str:
        for rule in self.rules:
            text = TextCleaningRule.registered_rules()[rule](text)
        return text


@TextCleaningRule
def strip_spaces(text: str) -> str:
    """Strips leading and trailing spaces/new lines"""
    return text.strip()


@TextCleaningRule
def rm_useless_spaces(text: str) -> str:
    """Removes multiple spaces in `str`"""
    return re.sub(" {2,}", " ", text)


@TextCleaningRule
def fix_html(text: str) -> str:
    """Replaces some special HTML characters: `&nbsp;`, `<br>`, etc."""
    text = (
        # non breakable space -> space
        text.replace("&nbsp;", " ")
        .replace("&#160;", " ")
        .replace("&#xa0;", " ")
        # <br> html single line breaks -> unicode line breaks
        .replace("<br>", "\n")
    )

    return text


@TextCleaningRule
def html_to_text(text: str) -> str:
    """Extracts text from an HTML document"""
    return BeautifulSoup(text, "lxml").get_text()



import copy
from typing import Any, Dict, List, Optional, Type, Union

from allennlp.common import FromParams, Params
from allennlp.data import TokenIndexer, Vocabulary
from allennlp.modules import TextFieldEmbedder

from biome.text.data import DataSource
from . import vocabulary
from .features import CharFeatures, WordFeatures
from .featurizer import InputFeaturizer
from .modules.encoders import Encoder
from .modules.heads.task_head import TaskHeadSpec
from .tokenizer import Tokenizer


class FeaturesConfiguration(FromParams):
    """Creates a input featurizer configuration
    
    This class will create a configuration for the features of the `Pipeline`.
    
    Use this for defining the main features to be used by the model, namely word and character embeddings.
    
    :::tip
    If you do not pass `words` and `chars` your pipeline will be setup with default word features (embedding_dim=50).
    :::
    
    Example:
    
    ```python
    word = WordFeatures(embedding_dim=100)
    char = CharFeatures(embedding_dim=16, encoder={'type': 'gru'})
    config = FeaturesConfiguration(word, char)
    ```
    
    Parameters
    ----------
    word : `biome.text.features.WordFeatures`
    char: `biome.text.features.CharFeatures`
    extra_params
    """

    __DEFAULT_CONFIG = WordFeatures(embedding_dim=50)

    def __init__(
        self,
        word: Optional[WordFeatures] = None,
        char: Optional[CharFeatures] = None,
        **extra_params
    ):
        self.word = word or None
        self.char = char or None

        for k, v in extra_params.items():
            self.__setattr__(k, v)

        if not (word or char or extra_params):
            self.word = self.__DEFAULT_CONFIG

    @classmethod
    def from_params(
        cls: Type["FeaturesConfiguration"], params: Params, **extras
    ) -> "FeaturesConfiguration":

        word = params.pop("word", params.pop("words", None))  # TODO: remove backward
        word = WordFeatures(**word.as_dict(quiet=True)) if word else None

        char = params.pop("char", params.pop("chars", None))  # TODO: remove backward
        char = CharFeatures(**char.as_dict(quiet=True)) if char else None

        return cls(word, char, **params.as_dict(), **extras)

    @property
    def keys(self) -> List[str]:
        """Gets the key features"""
        return [key for key in vars(self)]

    def compile_embedder(self, vocab: Vocabulary) -> TextFieldEmbedder:
        """Creates the embedder from configured features for a given vocabulary"""
        configuration = self._make_allennlp_config()
        if vocabulary.is_empty(vocab, namespaces=self.keys) and configuration.get(
            "word"
        ):
            # We simplify embedder configuration for better load an blank pipeline which create the vocab
            embedder_cfg = configuration["word"]["embedder"]
            embedder_cfg["embedding_dim"] = 1
            if "pretrained_file" in embedder_cfg:
                embedder_cfg["pretrained_file"] = None
        return TextFieldEmbedder.from_params(
            Params(
                {
                    "token_embedders": {
                        feature: config["embedder"]
                        for feature, config in configuration.items()
                    }
                }
            ),
            vocab=vocab,
        )

    def compile_featurizer(self, tokenizer: Tokenizer) -> InputFeaturizer:
        """Creates a featurizer from the configuration object
        
        :::tip
        
        If you are creating configurations programmatically use this method to check that your config object contains
        a valid configuration.
        
        :::

        Parameters
        ----------
        tokenizer: `Tokenizer`
            tokenizer used for this featurizer

        Returns
        -------
        The configured `InputFeaturizer`
        """
        configuration = self._make_allennlp_config()

        indexer = {
            feature: TokenIndexer.from_params(Params(config["indexer"]))
            for feature, config in configuration.items()
        }
        return InputFeaturizer(tokenizer, indexer=indexer)

    def _make_allennlp_config(self) -> Dict[str, Any]:
        """Creates compatible allennlp configuration"""
        configuration = {k: v for k, v in vars(self).items() if isinstance(v, dict)}
        configuration.update(
            {spec.namespace: spec.config for spec in [self.word, self.char] if spec}
        )
        return copy.deepcopy(configuration)


class TokenizerConfiguration(FromParams):
    """Creates a `Tokenizer` configuration

    Parameters
    ----------
    lang
    skip_empty_tokens
    max_sequence_length
    max_nr_of_sentences
    text_cleaning
    segment_sentences
    """

    def __init__(
        self,
        lang: str = "en",
        skip_empty_tokens: bool = False,
        max_sequence_length: int = None,
        max_nr_of_sentences: int = None,
        text_cleaning: Optional[Dict[str, Any]] = None,
        segment_sentences: Union[bool, Dict[str, Any]] = False,
    ):
        self.lang = lang
        self.skip_empty_tokens = skip_empty_tokens
        self.max_sequence_length = max_sequence_length
        self.max_nr_of_sentences = max_nr_of_sentences
        self.text_cleaning = text_cleaning
        self.segment_sentences = segment_sentences

    def compile(self) -> Tokenizer:
        """Build tokenizer object from its configuration"""
        return Tokenizer.from_params(Params(copy.deepcopy(vars(self))))


class PipelineConfiguration(FromParams):
    """"Creates a `Pipeline` configuration

    Parameters
    ----------
    name : `str`
        The `name` for our pipeline
    features : `FeaturesConfiguration`
        The input `features` to be used by the model pipeline. We define this using a `FeaturesConfiguration` object.
    head : `TaskHeadSpec`
        The `head` for the task, e.g., a LanguageModelling task, using a `TaskHeadSpec` object.
    tokenizer : `TokenizerConfiguration`, optional
        The `tokenizer` defined with a `TokenizerConfiguration` object.
    encoder : `Seq2SeqEncoderSpec`
        The core text seq2seq `encoder` of our model using a `Seq2SeqEncoderSpec`
    """

    def __init__(
        self,
        name: str,
        features: FeaturesConfiguration,
        head: TaskHeadSpec,
        tokenizer: Optional[TokenizerConfiguration] = None,
        encoder: Optional[Encoder] = None,
    ):
        super(PipelineConfiguration, self).__init__()

        self.name = name
        self.tokenizer = tokenizer or TokenizerConfiguration()
        self.features = features
        self.encoder = encoder
        self.head = head

    def as_dict(self) -> Dict[str, Any]:
        config = {
            "name": self.name,
            "tokenizer": vars(self.tokenizer),
            "features": vars(self.features),
            "head": self.head.config,
        }

        if self.encoder:
            config["encoder"] = self.encoder.config

        return config

    def build_tokenizer(self) -> Tokenizer:
        """Build the pipeline tokenizer"""
        return self.tokenizer.compile()

    def build_featurizer(self) -> InputFeaturizer:
        """Creates the pipeline featurizer"""
        return self.features.compile_featurizer(self.tokenizer.compile())

    def build_embedder(self, vocab: Vocabulary):
        """Build the pipeline embedder for aiven dictionary"""
        return self.features.compile_embedder(vocab)


class TrainerConfiguration:
    """ Creates a `TrainerConfiguration`

    Parameters
    ----------
    optimizer
    validation_metric
    patience
    num_epochs
    cuda_device
    grad_norm
    grad_clipping
    learning_rate_scheduler
    momentum_scheduler
    moving_average
    batch_size
    cache_instances
    in_memory_batches
    data_bucketing
    """

    def __init__(
        self,
        optimizer: Dict[str, Any] = None,
        validation_metric: str = "-loss",
        patience: Optional[int] = 2,
        num_epochs: int = 20,
        cuda_device: int = -1,
        grad_norm: Optional[float] = None,
        grad_clipping: Optional[float] = None,
        learning_rate_scheduler: Optional[Dict[str, Any]] = None,
        momentum_scheduler: Optional[Dict[str, Any]] = None,
        moving_average: Optional[Dict[str, Any]] = None,
        batch_size: Optional[int] = 16,
        cache_instances: bool = True,
        in_memory_batches: int = 2,
        data_bucketing: bool = True,
    ):
        self.optimizer = optimizer or {"type": "adam"}
        self.validation_metric = validation_metric
        self.patience = patience
        self.num_epochs = num_epochs
        self.cuda_device = cuda_device
        self.grad_norm = grad_norm
        self.grad_clipping = grad_clipping
        self.learning_rate_scheduler = learning_rate_scheduler
        self.momentum_scheduler = momentum_scheduler
        self.moving_average = moving_average

        # Data Iteration
        self.batch_size = batch_size
        self.data_bucketing = data_bucketing
        self.cache_instances = cache_instances
        self.in_memory_batches = in_memory_batches


class VocabularyConfiguration:
    """Configures a ``Vocabulary`` before it gets created from data

    Use this to configure a Vocabulary using specific arguments from `allennlp.data.Vocabulary``

    See [AllenNLP Vocabulary docs](https://docs.allennlp.org/master/api/data/vocabulary/#vocabulary])

    Parameters
    ----------
    sources : `List[DataSource]`
        Datasource to be used for data creation
    min_count : `Dict[str, int]`, optional (default=None)
        Minimum number of appearances of a token to be included in the vocabulary
    max_vocab_size :  `Union[int, Dict[str, int]]`, optional (default=`None`)
        Maximum number of tokens of the vocabulary
    pretrained_files : `Optional[Dict[str, str]]`, optional
        Pretrained files with word vectors
    only_include_pretrained_words : `bool`, optional (default=False)
        Only include tokens present in pretrained_files
    tokens_to_add : `Dict[str, int]`, optional
        A list of tokens to add to the vocabulary, even if they are not present in the ``sources``
    min_pretrained_embeddings : ``Dict[str, int]``, optional
        Minimum number of lines to keep from pretrained_files, even for tokens not appearing in the sources.
    """

    def __init__(
        self,
        sources: List[DataSource],
        min_count: Dict[str, int] = None,
        max_vocab_size: Union[int, Dict[str, int]] = None,
        pretrained_files: Optional[Dict[str, str]] = None,
        only_include_pretrained_words: bool = False,
        tokens_to_add: Dict[str, List[str]] = None,
        min_pretrained_embeddings: Dict[str, int] = None,
    ):
        self.sources = sources
        self.pretrained_files = pretrained_files
        self.min_count = min_count
        self.max_vocab_size = max_vocab_size
        self.only_include_pretrained_words = only_include_pretrained_words
        self.tokens_to_add = tokens_to_add
        self.min_pretrained_embeddings = min_pretrained_embeddings



import logging
import os.path
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import dask.dataframe as dd
import yaml
from biome.text.errors import MissingArgumentError

from .helpers import (
    is_relative_file_system_path,
    make_paths_relative,
)
from ..helpers import save_dict_as_yaml
from .readers import (
    ElasticsearchDataFrameReader,
    from_csv,
    from_excel,
    from_json,
    from_parquet,
)


class DataSource:
    """This class takes care of reading the data source, usually specified in a yaml file.

    It uses the *source readers* to extract a dask DataFrame.

    Parameters
    ----------
    source
        The data source. Could be a list of filesystem path, or a key name indicating the source backend (elasticsearch)
    format
        The data format. Optional. If found, overwrite the format extracted from source.
        Supported formats are listed as keys in the `SUPPORTED_FORMATS` dict of this class.
    mapping
        Used to map the features (columns) of the data source
        to the parameters of the DataSourceReader's `text_to_instance` method.
    **reader_options
        Additional kwargs are passed on to the *source readers* that depend on the format
        (see the `biome.text.data.readers` module).
    """

    _logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

    SUPPORTED_FORMATS = {
        "xls": (from_excel, dict(na_filter=False, keep_default_na=False, dtype=str)),
        "xlsx": (from_excel, dict(na_filter=False, keep_default_na=False, dtype=str)),
        "csv": (from_csv, dict(assume_missing=False, na_filter=False, dtype=str)),
        "json": (from_json, dict()),
        "jsonl": (from_json, dict()),
        "json-l": (from_json, dict()),
        "parquet": (from_parquet, dict()),
        # No file system based readers
        ElasticsearchDataFrameReader.SOURCE_TYPE: (
            ElasticsearchDataFrameReader.read,
            dict(),
        ),
    }
    # maps the supported formats to the corresponding "source readers"

    def __init__(
        self,
        source: Optional[Union[str, List[str]]] = None,
        mapping: Optional[Dict[str, Union[List[str], str]]] = None,
        format: Optional[str] = None,
        **reader_options,
    ):
        if not source:
            raise MissingArgumentError("source")

        self.source = source
        self.format = format or self.__format_from_source(source)
        self.mapping = mapping or {}
        self.reader_options = reader_options or {}

        source_reader, defaults = self._find_reader(self.format)

        data_frame = source_reader(source, **defaults, **self.reader_options)
        data_frame = self.__sanitize_dataframe(data_frame)
        # TODO allow disable index reindex
        if "id" in data_frame.columns:
            data_frame = data_frame.set_index("id")

        self._df = data_frame

    def __sanitize_dataframe(self, data_frame) -> dd.DataFrame:
        data_frame = data_frame.dropna(how="all")
        data_frame.columns = [
            column.strip() for column in data_frame.columns.astype(str).values
        ]
        for column in data_frame.columns:
            try:
                column_data = data_frame[column]
                if column_data.dtype.name == "category":
                    column_data = column_data.cat.add_categories("")
                data_frame[column] = column_data.fillna(value="")
            except ValueError as ex:
                self._logger.warning(
                    "Cannot set NaN's as empty string for column <%s>: %s", column, ex
                )
        return data_frame

    @classmethod
    def add_supported_format(
        cls, format_key: str, parser: Callable, default_params: Dict[str, Any] = None
    ) -> None:
        """Add a new format and reader to the data source readers.

        Parameters
        ----------
        format_key
            The new format key
        parser
            The parser function
        default_params
            Default parameters for the parser function
        """
        if format_key in cls.SUPPORTED_FORMATS.keys():
            cls._logger.warning("Already defined format %s", format_key)

        cls.SUPPORTED_FORMATS[format_key] = (parser, default_params or {})

    def to_dataframe(self) -> dd.DataFrame:
        """Returns the underlying DataFrame of the data source"""
        return self._df

    def to_mapped_dataframe(self) -> dd.DataFrame:
        """The columns of this DataFrame are named after the mapping keys, which in turn should match
        the parameter names in the DatasetReader's `text_to_instance` method.
        The content of these columns is specified in the mapping dictionary.

        Returns
        -------
        mapped_dataframe
            Contains columns corresponding to the parameter names of the DatasetReader's `text_to_instance` method.
        """
        if not self.mapping:
            raise ValueError("For a mapped DataFrame you need to specify a mapping!")

        # This is strictly a shallow copy of the underlying computational graph
        mapped_dataframe = self._df.copy()

        for parameter_name, data_features in self.mapping.items():
            # convert to list, otherwise the axis=1 raises an error with the returned pd.Series in the try statement
            # if no header is present in the source data, the column names are ints
            if isinstance(data_features, (str, int)):
                data_features = [data_features]

            try:
                mapped_dataframe[parameter_name] = self._df[data_features].apply(
                    self._to_dict_or_any, axis=1, meta=(None, "object")
                )
            except KeyError as err:
                raise KeyError(
                    err,
                    f"Did not find {data_features} in the data source columns {self._df.columns}!",
                )

        return mapped_dataframe[list(self.mapping.keys())]

    @staticmethod
    def _to_dict_or_any(value: dd.Series) -> Union[Dict, Any]:
        """Transform a `dask.dataframe.Series` to a dict or a single value, depending on its length."""
        if len(value) > 1:
            return value.to_dict()
        return value.iloc[0]

    @classmethod
    def from_yaml(cls: Type["DataSource"], file_path: str) -> "DataSource":
        """Create a data source from a yaml file.

        For the specific format, see the `self.to_yaml()` method.

        Parameters
        ----------
        file_path
            The path to the yaml file.

        Returns
        -------
        cls
        """
        with open(file_path) as yaml_file:
            cfg_dict = yaml.safe_load(yaml_file)

        # File system paths are usually specified relative to the yaml config file -> they have to be modified
        # 'path_keys' is not necessary, but specifying the dict keys
        # (for which we check for relative paths) is a safer choice
        make_paths_relative(os.path.dirname(file_path), cfg_dict, path_keys=["source"])

        cfg_dict.update(cfg_dict.pop("reader_options", {}))
        return cls(**cfg_dict)

    def to_yaml(self, path: str, make_source_path_absolute: bool = False) -> str:
        """Create a yaml config file for this data source.

        Parameters
        ----------
        path
            Path to the yaml file to be written.
        make_source_path_absolute
            If true, writes the source of the DataSource as an absolute path.

        Returns
        -------
        path
        """
        source = self.source
        if make_source_path_absolute and is_relative_file_system_path(source):
            source = os.path.abspath(source)

        yaml_dict = {
            "source": source,
            "format": self.format,
            "mapping": self.mapping,
            "reader_options": self.reader_options,
        }
        save_dict_as_yaml(yaml_dict, path)

        return path

    def head(self, n: int = 10) -> "pandas.DataFrame":  # pylint: disable=invalid-name
        """Allows for a peek into the data source showing the first n rows.

        Parameters
        ----------
        n
            Number of lines

        Returns
        -------
        df
            The first n lines as a `pandas.DataFrame`
        """
        return self._df.head(n=n)

    def _find_reader(self, source_format: str) -> Tuple[Callable, dict]:
        try:
            clean_format = source_format.lower().strip()
            return self.SUPPORTED_FORMATS[clean_format]
        except KeyError:
            raise TypeError(
                f"Format {source_format} not supported. Supported formats are: {', '.join(self.SUPPORTED_FORMATS)}"
            )

    @staticmethod
    def __format_from_source(source: Union[str, List[str]]) -> str:
        if isinstance(source, str):
            source = [source]
        formats = []
        for src in source:
            name, extension = os.path.splitext(src)
            formats.append(extension[1:] if extension else name)

        formats = set(formats)
        if len(formats) != 1:
            raise TypeError(f"source must be homogeneous: {formats}")
        return formats.pop()



import os

from biome.text.data import DataSource
from tests import DaskSupportTest, RESOURCES_PATH

FILES_PATH = os.path.join(RESOURCES_PATH, "data")


class JsonDatasourceTest(DaskSupportTest):
    def test_read_json(self):
        file_path = os.path.join(FILES_PATH, "dataset_source.jsonl")

        datasource = DataSource(format="json", source=file_path)
        data_frame = datasource.to_dataframe().compute()

        assert len(data_frame) > 0
        self.assertTrue("path" in data_frame.columns)

    def test_flatten_json(self):
        file_path = os.path.join(FILES_PATH, "to-be-flattened.jsonl")
        ds = DataSource(format="json", flatten=True, source=file_path)
        df = ds.to_dataframe().compute()

        for c in ["persons.*.lastName", "persons.*.name"]:
            self.assertIn(c, df.columns, f"Expected {c} as column name")

    def test_flatten_nested_list(self):
        file_path = os.path.join(FILES_PATH, "nested-list.jsonl")

        ds = DataSource(format="json", flatten=True, source=file_path)
        df = ds.to_dataframe().compute()

        for c in ["classification.*.origin.*.key", "classification.*.origin.*.source"]:
            self.assertIn(c, df.columns, f"Expected {c} as data column")



"""
    Manages vocabulary tasks and fetches vocabulary information

    Provides utilities for getting information from a given vocabulary.

    Provides management actions such as extending the labels, setting new labels or creating an "empty" vocab.
"""
import logging
from typing import Dict, List, Optional

from allennlp.data import Vocabulary
from allennlp.data.vocabulary import DEFAULT_OOV_TOKEN

LABELS_NAMESPACE = "gold_labels"

_LOGGER = logging.getLogger(__name__)


def get_labels(vocab: Vocabulary) -> List[str]:
    """Gets list of labels in the vocabulary

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`

    Returns
    -------
    labels: `List[str]`
        A list of label strings
    """
    return [k for k in vocab.get_token_to_index_vocabulary(namespace=LABELS_NAMESPACE)]


def label_for_index(vocab: Vocabulary, idx: int) -> str:
    """Gets label string for a label `int` id

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`
    idx: `int
        the token index

    Returns
    -------
    label: `str`
        The string for a label id
    """
    return vocab.get_token_from_index(idx, namespace=LABELS_NAMESPACE)


def index_for_label(vocab: Vocabulary, label: str) -> int:
    """Gets the label `int` id for label string

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary``
    label: `str`
        the label

    Returns
    -------
    label_idx: `int`
        The label id for label string
    """
    return vocab.get_token_index(label, namespace=LABELS_NAMESPACE)


def get_index_to_labels_dictionary(vocab: Vocabulary) -> Dict[int, str]:
    """Gets a dictionary for turning label `int` ids into label strings

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`

    Returns
    -------
    labels: `Dict[int, str]`
        A dictionary to get fetch label strings from ids
    """
    return vocab.get_index_to_token_vocabulary(LABELS_NAMESPACE)


def words_vocab_size(vocab: Vocabulary) -> int:
    """Fetches the vocabulary size for the `words` namespace

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`

    Returns
    -------
    size: `int`
        The vocabulary size for the words namespace
    """
    from biome.text.features import WordFeatures

    return vocab.get_vocab_size(WordFeatures.namespace)


def extend_labels(vocab: Vocabulary, labels: List[str]):
    """Adds a list of label strings to the vocabulary

    Use this to add new labels to your vocabulary (e.g., useful for reusing the weights of an existing classifier)

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`
    labels: `List[str]`
        A list of strings containing the labels to add to an existing vocabulary
    """
    vocab.add_tokens_to_namespace(labels, namespace=LABELS_NAMESPACE)


def set_labels(vocab: Vocabulary, new_labels: List[str]):
    """Resets the labels in the vocabulary with a given labels string list

    Parameters
    ----------
    vocab: `allennlp.data.Vocabulary`
    new_labels: `List[str]`
        The label strings to add to the vocabulary
    """
    for namespace_vocab in [
        vocab.get_token_to_index_vocabulary(LABELS_NAMESPACE),
        vocab.get_index_to_token_vocabulary(LABELS_NAMESPACE),
    ]:
        tokens = list(namespace_vocab.keys())
        for token in tokens:
            del namespace_vocab[token]

    extend_labels(vocab, new_labels)


def empty_vocabulary(namespaces):
    """
    Creates an empty vocabulary. Used for early pipeline initialization

    Arguments
    ----------
    namespaces: `List[str]`
        The vocab namespaces to create
    """
    vocab = Vocabulary()
    for namespace in namespaces:
        vocab.add_token_to_namespace(DEFAULT_OOV_TOKEN, namespace=namespace)
    return vocab


def is_empty(vocab: Vocabulary, namespaces: List[str]) -> bool:
    """
    Checks if a vocab is empty respect to given namespaces

    Returns True vocab size is 0 for all given namespaces
    """
    for namespaces in namespaces:
        # We must drop the padding and out of vocab tokens = 2 tokens
        if vocab.get_vocab_size(namespaces) > 2:
            return False
    return True


def load_vocabulary(vocab_path: str) -> Optional[Vocabulary]:
    """
    Loads a vocabulary from a path
    Parameters
    ----------
    vocab_path: str
        The vocab folder path

    Returns
    -------
    An operative `allennlp.data.Vocabulary`
    """
    try:
        if not vocab_path:
            return None
        return Vocabulary.from_files(vocab_path)
    except TypeError:
        return None
    except FileNotFoundError:
        _LOGGER.warning("%s folder not found", vocab_path)
        return None



import logging
import os

from gevent.pywsgi import WSGIServer

from biome.text.environment import ES_HOST
from .app import make_app

# TODO centralize configuration
logging.basicConfig(level=logging.INFO)

__LOGGER = logging.getLogger(__name__)

STATICS_DIR = os.path.join(os.path.dirname(__file__), "webapp")


def launch_ui(es_host: str, port: int = 9000) -> None:
    es_host = es_host if es_host else os.getenv(ES_HOST, "http://localhost:9200")

    flask_app = make_app(
        es_host=es_host,
        statics_dir=STATICS_DIR
        # statics_dir=temporal_static_path("classifier")
    )

    http_server = WSGIServer(("0.0.0.0", port), flask_app)
    __LOGGER.info(
        "Running biome UI on %s with elasticsearch backend %s",
        f"http://0.0.0.0:{http_server.server_port}",
        es_host,
    )
    http_server.serve_forever()



import inspect
import json
import logging
import os
import pathlib
import pickle
import re
import warnings
from copy import deepcopy
from functools import lru_cache
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, Union, cast

import allennlp
import numpy
import torch
from allennlp.common import Params
from allennlp.common.util import prepare_environment, sanitize
from allennlp.data import DataLoader, DatasetReader, Field, Instance, Vocabulary
from allennlp.data.fields import ListField, TextField
from allennlp.models.archival import CONFIG_NAME, archive_model
from allennlp.training import Trainer
from allennlp.training.util import evaluate
from dask.dataframe import Series as DaskSeries
from torch.utils.data.dataset import Dataset

from biome.text import vocabulary
from .backbone import ModelBackbone
from .configuration import PipelineConfiguration
from .data import DataSource
from .errors import MissingArgumentError, WrongValueError
from .features import WordFeatures
from .helpers import split_signature_params_by_predicate
from .modules.heads import TaskHead, TaskOutput


class _HashDict(dict):
    """
    Hashable dict implementation.
    BE AWARE! Since dicts are mutable, the hash can change!
    """

    def __hash__(self):
        # user a better way
        return pickle.dumps(self).__hash__()


class _HashList(list):
    """
    Hashable list implementation.
    BE AWARE! Since dicts are mutable, the hash can change!
    """

    def __hash__(self):
        return pickle.dumps(self).__hash__()


class PipelineModel(allennlp.models.Model, allennlp.data.DatasetReader):
    """
    This class represents pipeline model implementation for connect biome.text concepts with
    allennlp implementation details

    This class manage the head + backbone encoder, keeping the allennlnlp Model lifecycle. This class
    should be hidden to api users.
    """

    def __init__(self, name: str, head: TaskHead):
        allennlp.models.Model.__init__(self, head.backbone.vocab)
        allennlp.data.DatasetReader.__init__(self, lazy=True)

        self._head = None
        self.name = name
        self.set_head(head)

    def _update_head_related_attributes(self):
        """Updates the inputs/outputs and default mapping attributes, calculated from model head"""
        required, optional = split_signature_params_by_predicate(
            self._head.featurize, lambda p: p.default == inspect.Parameter.empty
        )
        self._inputs = self._head.inputs() or [p.name for p in required]
        self._output = ([p.name for p in optional] or [None])[0]
        self._default_ds_mapping = {k: k for k in self._inputs + [self._output] if k}

    @classmethod
    def from_params(
        cls: Type["PipelineModel"],
        params: Params,
        vocab: Optional[Vocabulary] = None,
        **extras,
    ) -> "PipelineModel":
        """
        Load the model implementation from params. We build manually each component from config sections.

        The param keys matches exactly with keys in yaml configuration files
        """

        config = params.pop("config")
        if not isinstance(config, PipelineConfiguration):
            config = PipelineConfiguration.from_params(config)

        vocab = vocab or vocabulary.empty_vocabulary(namespaces=config.features.keys)
        tokenizer = config.tokenizer.compile()
        featurizer = config.features.compile_featurizer(tokenizer)
        embedder = config.build_embedder(vocab)

        return cls(
            name=config.name,
            head=config.head.compile(
                backbone=ModelBackbone(
                    vocab,
                    featurizer=featurizer,
                    embedder=embedder,
                    encoder=config.encoder,
                )
            ),
        )

    @property
    def head(self) -> TaskHead:
        """Get the model head"""
        return self._head

    def set_head(self, head: TaskHead) -> None:
        """Set a head and update related model attributes"""
        self._head = head
        self._update_head_related_attributes()

    def cache_data(self, cache_directory: str) -> None:
        """Sets the cache data directory"""
        self._cache_directory = pathlib.Path(cache_directory)
        os.makedirs(self._cache_directory, exist_ok=True)

    def forward(self, *args, **kwargs) -> Dict[str, torch.Tensor]:
        """The main forward method. Wraps the head forward method and converts the head output into a dictionary"""
        head_output: TaskOutput = self._head.forward(*args, **kwargs)
        # we don't want to break AllenNLP API: TaskOutput -> as_dict()
        return head_output.as_dict()

    def make_output_human_readable(
        self, output_dict: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Completes the output for the prediction

        Parameters
        ----------
        output_dict
            The `TaskOutput` from the model's forward method as dict

        Returns
        -------
        output_dict
            Completed output
        """
        # we don't want to break AllenNLP API: dict -> TaskOutput -> dict
        output = TaskOutput(**output_dict)
        completed_output = self._head.decode(output)
        return completed_output.as_dict()

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        """Fetch metrics defined in head layer"""
        return self._head.get_metrics(reset)

    def text_to_instance(self, **inputs: Dict[str, Any]) -> Optional[Instance]:
        """Applies the head featurize method"""
        try:
            return self._head.featurize(**inputs)
        except KeyError as error:
            # missing inputs
            raise MissingArgumentError(arg_name=error.args[0])

    def set_vocab(self, vocab: Vocabulary):
        """Replace the current vocab and reload all model layer"""
        self.vocab = vocab
        self._head.backbone.vocab = vocab
        self._head.backbone.on_vocab_update()
        self._head.on_vocab_update()

    @property
    def inputs(self) -> List[str]:
        """The model inputs. Corresponding to head.featurize required argument names"""
        return self._inputs

    @property
    def output(self) -> Optional[str]:
        """The model output (not prediction): Corresponding to head.featurize optional argument names. Should be one"""
        return self._output

    def init_prediction_logger(
        self, output_dir: str, max_bytes: int = 20000000, backup_count: int = 20
    ):
        """Initialize the prediction logger.

        If initialized we will log all predictions to a file called *predictions.json* in the `output_folder`.

        Parameters
        ----------
        output_dir
            Path to the folder in which we create the *predictions.json* file.
        max_bytes
            Passed on to logging.handlers.RotatingFileHandler
        backup_count
            Passed on to logging.handlers.RotatingFileHandler

        """
        Path(output_dir).mkdir(parents=True, exist_ok=True)

        predictions_logger = logging.getLogger(output_dir)
        predictions_logger.setLevel(logging.DEBUG)
        # This flag avoids logging messages to be propagated to the parent loggers
        predictions_logger.propagate = False
        file_handler = RotatingFileHandler(
            os.path.join(output_dir, self.PREDICTION_FILE_NAME),
            maxBytes=max_bytes,
            backupCount=backup_count,
        )
        file_handler.setLevel(logging.INFO)
        predictions_logger.addHandler(file_handler)

        setattr(self, "_prediction_logger", predictions_logger)

    def init_prediction_cache(self, max_size: int) -> None:
        """Initialize a prediction cache using the functools.lru_cache decorator

        Parameters
        ----------
        max_size
            Save up to max_size most recent items.
        """

        if hasattr(self, "_predict_with_cache"):
            warnings.warn(
                "Prediction cache already initiated!", category=RuntimeWarning
            )
            return

        predict_with_cache = lru_cache(maxsize=max_size)(self.predict)

        def predict_wrapper(*args, **kwargs):
            def hashable_value(value) -> Any:
                if isinstance(value, dict):
                    return _HashDict(value)
                if isinstance(value, (list, tuple)):
                    return _HashList(value)
                return value

            return predict_with_cache(
                *[hashable_value(arg_value) for arg_value in args],
                **{
                    key: hashable_value(input_value)
                    for key, input_value in kwargs.items()
                },
            )

        self.__setattr__("predict", predict_wrapper)
        self.__setattr__("_predict_with_cache", predict_with_cache)

    def log_prediction(
        self, inputs: Dict[str, Any], prediction: Dict[str, Any]
    ) -> None:
        """Store prediction for model analysis and feedback sessions"""
        if hasattr(self, "_prediction_logger"):
            self._prediction_logger.info(
                json.dumps(dict(inputs=inputs, annotation=sanitize(prediction)))
            )

    def predict(self, *args, **kwargs) -> Dict[str, numpy.ndarray]:
        """Make a single model prediction"""
        if self.training:
            warnings.warn(
                "Train model enabled. "
                "Disabling training mode automatically. You can manually disable it: "
                "self.eval() or self.train(False)"
            )
            self.eval()

        inputs = self._model_inputs_from_args(*args, **kwargs)
        instance = self.text_to_instance(**inputs)
        prediction = self.forward_on_instance(instance)
        self.log_prediction(inputs, prediction)

        return prediction

    def explain(self, *args, **kwargs) -> Dict[str, Any]:
        """Applies an prediction including token attribution explanation"""
        inputs = self._model_inputs_from_args(*args, **kwargs)
        instance = self.text_to_instance(**inputs)
        tokenization = self._get_instance_tokenization(instance)

        prediction = self.forward_on_instance(instance)
        explain = self._head.explain_prediction(
            prediction=prediction, instance=instance
        )
        # TODO: change explain data model include both, tokenization and attributions information
        #  We must apply this change in sync with frontend team.
        #  The data model proposal could be as follow:
        #    {
        #        "tokenization": { "text": ["This", "is", "an", "example"] },
        #        "attributions": { "text": [{"token": "This", "attribution": 0.39},...]}
        #    }
        #   For now, we add tokenization as a new separate field
        explain.update({"tokenization": tokenization})

        return explain

    def _get_instance_tokenization(self, instance: Instance) -> Dict[str, Any]:
        """Gets the tokenization information to current instance"""

        def extract_field_tokens(field: Field) -> Union[List[str], List[List[str]]]:
            """Tries to extract tokens from field"""
            if isinstance(field, TextField):
                return [token.text for token in cast(TextField, field).tokens]
            if isinstance(field, ListField):
                return [
                    extract_field_tokens(inner_field)
                    for inner_field in cast(ListField, field)
                ]
            raise WrongValueError(f"Cannot extract fields from [{type(field)}]")

        return {name: extract_field_tokens(field) for name, field in instance.items()}

    def _model_inputs_from_args(self, *args, **kwargs) -> Dict[str, Any]:
        """Returns model input data dictionary"""
        inputs = {k: v for k, v in zip(self.inputs, args)}
        inputs.update(kwargs)

        return inputs

    def launch_experiment(
        self,
        params: Params,
        serialization_dir: str,
        batch_weight_key: str = "",
        embedding_sources_mapping: Dict[str, str] = None,
    ) -> "PipelineModel":
        """Launch a local experiment for model training"""

        trainer = PipelineModelTrainer(
            self,
            params,
            serialization_dir,
            batch_weight_key,
            embedding_sources_mapping,
        )

        model, _ = trainer.train()
        return model

    def _read(self, file_path: str) -> Iterable[Instance]:
        """A generator that yields `Instance`s that are fed to the model

        This method is implicitly called when training the model.

        Parameters
        ----------
        file_path
            Path to the configuration file (yml) of the data source.

        Yields
        ------
        instance
            An `Instance` that is fed to the model
        """
        data_source = DataSource.from_yaml(file_path)
        if not data_source.mapping:
            data_source.mapping = self._default_ds_mapping

        dataframe = data_source.to_mapped_dataframe()
        instances: DaskSeries = dataframe.apply(
            lambda x: self.text_to_instance(**x.to_dict()),
            axis=1,
            meta=(None, "object"),
        )

        return (instance for _, instance in instances.iteritems() if instance)


class PipelineModelTrainer:
    """
    Default trainer for ``PipelineModel``

    Arguments
    ----------
    model : ``PipelineModel``
        The trainable model
    params : ``Params``
        A parameter object specifying an AllenNLP Experiment
    serialization_dir : ``str``
        The directory in which to save results and logs.
    batch_weight_key : ``str``, optional (default="")
        If non-empty, name of metric used to weight the loss on a per-batch basis.
    embedding_sources_mapping: ``Dict[str, str]``, optional (default=None)
        mapping from model paths to the pretrained embedding filepaths
        used during fine-tuning.
    """

    __LOGGER = logging.getLogger(__name__)

    def __init__(
        self,
        model: PipelineModel,
        params: Params,
        serialization_dir: str,
        batch_weight_key: str = "",
        embedding_sources_mapping: Dict[str, str] = None,
    ):
        self._model = model
        # TODO use biome classes instead of common parameters
        self._params = params
        self._serialization_dir = serialization_dir
        self._batch_weight_key = batch_weight_key
        self._embedding_sources_mapping = embedding_sources_mapping
        self._batch_size = self._params.pop("iterator")["batch_size"]

        self._trainer = None
        self._all_datasets = self.datasets_from_params()

        self._setup()

    def _setup(self):
        """Setup the trainer components and local resources"""

        prepare_environment(self._params)
        if os.path.exists(self._serialization_dir) and os.listdir(
            self._serialization_dir
        ):
            self.__LOGGER.info(
                f"Serialization directory ({self._serialization_dir}) "
                f"already exists and is not empty."
            )

        os.makedirs(self._serialization_dir, exist_ok=True)

        # We don't need to load pretrained weights from saved models
        try:
            self._params["model"]["config"]["features"]["word"].weights_file = None
        # there is no word feature
        except AttributeError:
            pass

        serialization_params = sanitize(deepcopy(self._params).as_dict(quiet=True))
        with open(
            os.path.join(self._serialization_dir, CONFIG_NAME), "w"
        ) as param_file:
            json.dump(serialization_params, param_file, indent=4)

        self._params.pop("model", None)

        vocab = self._model.vocab
        vocab.save_to_files(os.path.join(self._serialization_dir, "vocabulary"))

        for dataset in self._all_datasets.values():
            if dataset and hasattr(dataset, "index_with"):
                dataset.index_with(vocab)

        trainer_params = self._params.pop("trainer")
        no_grad_regexes = trainer_params.pop(
            "no_grad", ()
        )  # This could be nice to have exposed
        for name, parameter in self._model.named_parameters():
            if any(re.search(regex, name) for regex in no_grad_regexes):
                parameter.requires_grad_(False)

        train_data_loader = DataLoader(
            self._all_datasets["train"], batch_size=self._batch_size
        )
        validation_ds = self._all_datasets.get("validation")
        validation_data_loader = (
            DataLoader(validation_ds, batch_size=self._batch_size)
            if validation_ds
            else None
        )

        # TODO: Customize trainer for better biome integration
        self._trainer = Trainer.from_params(
            model=self._model,
            serialization_dir=self._serialization_dir,
            data_loader=train_data_loader,
            validation_data_loader=validation_data_loader,
            params=trainer_params,
        )

    def datasets_from_params(self) -> Dict[str, Dataset]:
        """
        Load all the datasets specified by the config.

        """

        self._params.pop("dataset_reader")
        self._params.pop("validation_dataset_reader", None)
        train_data_path = self._params.pop("train_data_path")

        self.__LOGGER.info("Reading training data from %s", train_data_path)
        datasets: Dict[str, Dataset] = {"train": self._model.read(train_data_path)}

        validation_data_path = self._params.pop("validation_data_path", None)
        if validation_data_path is not None:
            self.__LOGGER.info("Reading validation data from %s", validation_data_path)
            datasets["validation"] = self._model.read(validation_data_path)

        test_data_path = self._params.pop("test_data_path", None)
        if test_data_path is not None:
            self.__LOGGER.info("Reading test data from %s", test_data_path)
            datasets["test"] = self._model.read(test_data_path)

        return datasets

    def test_evaluation(self) -> Dict[str, Any]:
        """
        Evaluates the model against the test dataset (if defined)

        Returns
        -------
        Test metrics information

        """
        test_data = self._all_datasets.get("test")
        if not test_data:
            return {}

        self.__LOGGER.info("The model will be evaluated using the best epoch weights.")
        return evaluate(
            self._model,
            data_loader=DataLoader(test_data, batch_size=self._batch_size),
            cuda_device=self._trainer.cuda_device,
            batch_weight_key=self._batch_weight_key,
        )

    def train(self) -> Tuple[PipelineModel, Dict[str, Any]]:
        """
        Train the inner model with given configuration on initialization

        Returns
        -------
        A tuple with trained model and related metrics information
        """

        from allennlp.models.model import _DEFAULT_WEIGHTS

        try:
            metrics = self._trainer.train()
        except KeyboardInterrupt:
            # if we have completed an epoch, try to create a model archive.
            if os.path.exists(os.path.join(self._serialization_dir, _DEFAULT_WEIGHTS)):
                logging.info(
                    "Fine-tuning interrupted by the user. Attempting to create "
                    "a model archive using the current best epoch weights."
                )
                self.save_best_model()
            raise

        for k, v in self.test_evaluation().items():
            metrics["test_" + k] = v

        self.save_best_model()

        with open(
            os.path.join(self._serialization_dir, "metrics.json"), "w"
        ) as metrics_file:
            metrics_json = json.dumps(metrics, indent=2)
            metrics_file.write(metrics_json)

        return self._model, metrics

    def save_best_model(self):
        """Packages the best model as tar.gz archive"""
        archive_model(self._serialization_dir)



from biome.text.data import DataSource
import os

from tests import DaskSupportTest, RESOURCES_PATH

FILES_PATH = os.path.join(RESOURCES_PATH, "data")


class CsvDatasourceTest(DaskSupportTest):
    def test_read_csv(self):
        file_path = os.path.join(FILES_PATH, "dataset_source.csv")

        datasource = DataSource(format="csv", source=file_path)
        data_frame = datasource.to_dataframe().compute()

        assert len(data_frame) > 0
        self.assertTrue("path" in data_frame.columns)

    def test_reader_csv_with_leading_and_trailing_spaces_in_examples(self):
        ds = DataSource(
            format="csv",
            source=os.path.join(FILES_PATH, "trailing_coma_in_headers.csv"),
            sep=";",
        )
        df = ds.to_dataframe().compute()
        self.assertIn("name", df.columns)



from allennlp.modules import BiMpmMatching, FeedForward, Seq2SeqEncoder, Seq2VecEncoder

from .defs import ComponentSpec


class Seq2VecEncoderSpec(ComponentSpec[Seq2VecEncoder]):
    """Layer spec for Seq2VecEncoder components"""

    pass


class Seq2SeqEncoderSpec(ComponentSpec[Seq2SeqEncoder]):
    """Layer spec for Seq2SeqEncoder components"""

    pass


class FeedForwardSpec(ComponentSpec[FeedForward]):
    """Layer spec for FeedForward components"""

    pass


class BiMpmMatchingSpec(ComponentSpec[BiMpmMatching]):
    """Layer spec for BiMpmMatching components"""

    pass



import warnings
from typing import Any, Dict, List, Optional, Union

import torch
from allennlp.data import Instance, TextFieldTensors, Vocabulary
from allennlp.modules import TextFieldEmbedder
from allennlp.modules.seq2seq_encoders import PassThroughEncoder

from .featurizer import InputFeaturizer
from .modules.encoders import Encoder


class ModelBackbone(torch.nn.Module):
    """The backbone of the model.

    It is composed of a tokenizer, featurizer and an encoder.
    This component of the model can be pretrained and used with different task heads.

    Attributes
    ----------
    vocab : `Vocabulary`
        The vocabulary of the pipeline
    featurizer : `InputFeaturizer`
        Defines the input features of the tokens and indexes
    embedder: `TextFieldEmbedder`
        The backbone embedder layer
    encoder : Encoder
        Outputs an encoded sequence of the tokens
    """

    def __init__(
        self,
        vocab: Vocabulary,
        featurizer: InputFeaturizer,
        embedder: TextFieldEmbedder,
        encoder: Optional[Encoder] = None,
    ):
        super(ModelBackbone, self).__init__()

        self.vocab = vocab
        self.featurizer = featurizer
        self.embedder = embedder
        self.encoder = (
            encoder.input_dim(self.embedder.get_output_dim()).compile()
            if encoder
            else PassThroughEncoder(self.embedder.get_output_dim())
        )

    def forward(
        self, text: TextFieldTensors, mask: torch.Tensor, num_wrapping_dims: int = 0,
    ) -> torch.Tensor:
        """Applies embedding + encoder layers"""
        embeddings = self.embedder(text, num_wrapping_dims=num_wrapping_dims)
        return self.encoder(embeddings, mask=mask)

    def on_vocab_update(self):
        """This method is called when a base model updates the vocabulary"""
        # This loop applies only for embedding layer.
        for model_path, module in self.named_modules():
            if hasattr(module, "extend_vocab"):
                module.extend_vocab(self.vocab)

    def featurize(
        self,
        record: Union[str, List[str], Dict[str, Any]],
        to_field: str = "record",
        aggregate: bool = False,
        tokenize: bool = True,
    ) -> Instance:
        """
        Generate a allennlp Instance from a record input.

        If aggregate flag is enabled, the resultant instance will contains a single TextField's
        with all record fields; otherwhise, a ListField of TextFields.

        Parameters
        ----------
        record: `Union[str, List[str], Dict[str, Any]]`
            input data
        to_field: `str`
            field name in returned instance
        aggregate: `bool`
            set data aggregation flag
        tokenize: `bool`
            If disabled, skip tokenization phase, and pass record data as tokenized token list.

        Returns
        -------

        instance: `Instance`

        Deprecated: use self.featurizer instead

        """
        warnings.warn(
            "backbone.featurize is deprecated. Use instead backbone.featurizer",
            DeprecationWarning,
        )
        return self.featurizer(record, to_field, aggregate, tokenize)



from typing import Dict, List, Optional, Union, cast

import torch
from allennlp.data import Instance, TextFieldTensors
from allennlp.data.fields import SequenceLabelField, TextField
from allennlp.modules import ConditionalRandomField, FeedForward, TimeDistributed
from allennlp.modules.conditional_random_field import allowed_transitions
from allennlp.nn.util import get_text_field_mask
from allennlp.training.metrics import CategoricalAccuracy, SpanBasedF1Measure

from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.modules.specs import ComponentSpec, FeedForwardSpec
from .task_head import TaskHead, TaskName, TaskOutput


class TokenClassification(TaskHead):
    """Task head for token classification (NER)"""

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        label_encoding: Optional[str] = "BIOUL",
        dropout: Optional[float] = None,
        feedforward: Optional[FeedForwardSpec] = None,
    ) -> None:
        super(TokenClassification, self).__init__(backbone)
        vocabulary.set_labels(self.backbone.vocab, labels)

        if dropout:
            self.dropout = torch.nn.Dropout(dropout)
        else:
            self.dropout = None

        self._feedforward: FeedForward = (
            None
            if not feedforward
            else feedforward.input_dim(backbone.encoder.get_output_dim()).compile()
        )
        # output layers
        self._classifier_input_dim = (
            self._feedforward.get_output_dim()
            if self._feedforward
            else backbone.encoder.get_output_dim()
        )
        # we want this linear applied to each token in the sequence
        self._label_projection_layer = TimeDistributed(
            torch.nn.Linear(self._classifier_input_dim, self.num_labels)
        )
        constraints = allowed_transitions(
            label_encoding,
            vocabulary.get_index_to_labels_dictionary(self.backbone.vocab),
        )
        self._crf = ConditionalRandomField(
            self.num_labels, constraints, include_start_end_transitions=True
        )
        self._metrics = {
            "accuracy": CategoricalAccuracy(),
            "f1": SpanBasedF1Measure(
                self.backbone.vocab,
                tag_namespace=vocabulary.LABELS_NAMESPACE,
                label_encoding=label_encoding,
            ),
        }
        # loss is calculated as -log_likelihood from crf
        self._loss = lambda logits, label, mask: -(self._crf(logits, label, mask))

    def featurize(
        self, text: List[str], labels: Optional[Union[List[str], List[int]]] = None
    ) -> Optional[Instance]:

        instance = self.backbone.featurizer(
            text, to_field="text", tokenize=False, aggregate=True
        )

        if labels:
            instance.add_field(
                "labels",
                SequenceLabelField(
                    labels,
                    sequence_field=cast(TextField, instance["text"]),
                    label_namespace=vocabulary.LABELS_NAMESPACE,
                ),
            )

        return instance

    def task_name(self) -> TaskName:
        return TaskName.token_classification

    def forward(  # type: ignore
        self, text: TextFieldTensors, labels: torch.IntTensor = None
    ) -> TaskOutput:
        mask = get_text_field_mask(text)
        embedded_text = self.backbone.forward(text, mask)

        if self.dropout:
            encoded_text = self.dropout(embedded_text)

        if self._feedforward is not None:
            embedded_text = self._feedforward(embedded_text)

        logits = self._label_projection_layer(embedded_text)
        best_paths = self._crf.viterbi_tags(logits, mask)
        predicted_tags = cast(List[List[int]], [x[0] for x in best_paths])
        class_probabilities = logits * 0.0

        for i, instance_tags in enumerate(predicted_tags):
            for j, tag_id in enumerate(instance_tags):
                class_probabilities[i, j, tag_id] = 1

        output = TaskOutput(
            logits=logits,
            probs=class_probabilities,
            mask=mask,
            tags=[
                [vocabulary.label_for_index(self.backbone.vocab, idx) for idx in tags]
                for tags in predicted_tags
            ],
        )

        if labels is not None:
            output.loss = self._loss(logits, labels, mask)
            for metric in self._metrics.values():
                metric(class_probabilities, labels, mask)

        return output

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        metrics = {
            metric_name: metric.get_metric(reset)
            for metric_name, metric in self._metrics.items()
            if metric_name != "f1"
        }
        f1_dict = self._metrics.get("f1").get_metric(reset=reset)
        metrics.update(f1_dict)
        return metrics


class TokenClassificationSpec(ComponentSpec[TokenClassification]):
    """Spec for classification head components"""

    pass



import os
from typing import Optional

import click
from click import Path

from biome.text import Pipeline, TrainerConfiguration, VocabularyConfiguration
from biome.text.data import DataSource
from biome.text.helpers import yaml_to_dict


@click.command("train", help="Train a pipeline")
@click.argument("pipeline_path", type=Path(exists=True))
@click.option("-o", "--output", "output", type=Path(), required=True)
@click.option("--trainer", "trainer", type=Path(exists=True), required=True)
@click.option("--training", "training", type=Path(exists=True), required=True)
@click.option("--validation", "validation", type=Path(exists=True), required=False)
@click.option("--test", "test", type=Path(exists=True), required=False)
def learn(
    pipeline_path: str,
    output: str,
    trainer: str,
    training: str,
    validation: Optional[str] = None,
    test: Optional[str] = None,
) -> None:
    _, extension = os.path.splitext(pipeline_path)
    extension = extension[1:].lower()
    pipeline = (
        Pipeline.from_yaml(pipeline_path)
        if extension in ["yaml", "yml"]
        else Pipeline.from_pretrained(pipeline_path)
    )

    pipeline.create_vocabulary(
        VocabularyConfiguration(
            sources=[
                DataSource.from_yaml(ds) for ds in [training, validation, test] if ds
            ]
        ),
    )

    pipeline.train(
        output=output,
        trainer=TrainerConfiguration(**yaml_to_dict(trainer)),
        training=training,
        validation=validation,
        test=test,
    )



import json
import os
from logging import Logger

import requests
from flask import (
    Flask,
    request,
    Response,
    send_file,
    send_from_directory,
    jsonify,
    redirect,
)
from flask_cors import CORS
from werkzeug.exceptions import NotFound
from werkzeug.middleware.proxy_fix import ProxyFix


def make_app(es_host: str, statics_dir: str) -> Flask:
    app = Flask(__name__)  # pylint: disable=invalid-name
    app.wsgi_app = ProxyFix(
        app.wsgi_app
    )  # sets the requester IP with the X-Forwarded-For header
    CORS(app)

    logger: Logger = app.logger

    @app.route("/elastic/<path:es_path>", methods=["GET", "OPTIONS"])
    def es_info_proxy(es_path: str) -> Response:
        es_url = f"{es_host}/{es_path}"

        if request.method == "OPTIONS":
            return Response(response="", status=200)

        response = requests.get(es_url)
        return jsonify(response.json())

    @app.route("/elastic/<path:es_path>/_update", methods=["POST", "OPTIONS"])
    def es_update_proxy(es_path: str) -> Response:
        es_url = f"{es_host}/{es_path}/_update"
        response = requests.post(es_url, json=request.get_json())
        return jsonify(response.json())

    @app.route("/elastic/<path:index>/_search", methods=["GET", "POST", "OPTIONS"])
    def es_search_proxy(
        index: str = None,
    ) -> Response:  # pylint: disable=unused-variable
        if request.method == "OPTIONS":
            return Response(response="", status=200)

        es_url = (
            "{}/{}/_search".format(es_host, index)
            if index
            else "{}/_search".format(es_host)
        )

        if request.method == "GET":
            response = requests.get(es_url)
        else:
            response = requests.post(es_url, json=request.get_json())

        return jsonify(json.loads(response.text))

    @app.route("/<path:path>")
    def static_proxy(path: str) -> Response:  # pylint: disable=unused-variable
        try:
            return send_from_directory(statics_dir, path)
        except NotFound as error:
            logger.warning(error)
            return index()

    @app.route("/static/js/<path:path>")
    def static_js_proxy(path: str) -> Response:  # pylint: disable=unused-variable
        return send_from_directory(os.path.join(statics_dir, "static/js"), path)

    @app.route("/static/css/<path:path>")
    def static_css_proxy(path: str) -> Response:  # pylint: disable=unused-variable
        return send_from_directory(os.path.join(statics_dir, "static/css"), path)

    @app.route("/public/<path:path>")
    def public_proxy(path: str) -> Response:  # pylint: disable=unused-variable
        return send_from_directory(statics_dir, path)

    @app.route("/")
    def index() -> Response:  # pylint: disable=unused-variable
        return send_file(os.path.join(statics_dir, "index.html"))

    return app



import datetime
from typing import Any, Dict, Optional

import allennlp
from biome.text import constants
from biome.text._model import PipelineModel
from elasticsearch import Elasticsearch

from . import helpers
from .configuration import TrainerConfiguration

_ModelImpl = PipelineModel


def __register(impl_class, overrides: bool = False):
    """Register the impl. class in allennlp components registry"""

    allennlp.models.Model.register(impl_class.__name__, exist_ok=overrides)(impl_class)
    allennlp.data.DatasetReader.register(impl_class.__name__, exist_ok=overrides)(
        impl_class
    )


__register(_ModelImpl, overrides=True)


class ExploreConfiguration:
    """Configures an exploration run

    Parameters
    ----------
        batch_size: `int`
            The batch size for indexing predictions (default is `500)
        prediction_cache_size: `int`
            The size of the cache for caching predictions (default is `0)
        explain: `bool`
            Whether to extract and return explanations of token importance (default is `False`)
        force_delete: `bool`
            Whether to delete existing explore with `explore_id` before indexing new items (default is `True)
        metadata: `kwargs`
            Additional metadata to index in Elasticsearch
    """

    def __init__(
        self,
        batch_size: int = 500,
        prediction_cache_size: int = 0,
        explain: bool = False,
        force_delete: bool = True,
        **metadata,
    ):
        self.batch_size = batch_size
        self.prediction_cache = prediction_cache_size
        self.explain = explain
        self.force_delete = force_delete
        self.metadata = metadata


class ElasticsearchExplore:
    """Elasticsearch data exploration class"""

    def __init__(self, es_index: str, es_host: Optional[str] = None):
        self.es_index = es_index
        self.es_host = es_host or constants.DEFAULT_ES_HOST
        if not self.es_host.startswith("http"):
            self.es_host = f"http://{self.es_host}"

        self.client = Elasticsearch(
            hosts=es_host, retry_on_timeout=True, http_compress=True
        )
        self.es_doc = helpers.get_compatible_doc_type(self.client)

    def create_explore_data_record(self, parameters: Dict[str, Any]):
        """Creates an exploration data record data exploration"""

        self.client.indices.create(
            index=constants.BIOME_METADATA_INDEX,
            body={
                "settings": {"index": {"number_of_shards": 1, "number_of_replicas": 0}}
            },
            params=dict(ignore=400),
        )

        self.client.update(
            index=constants.BIOME_METADATA_INDEX,
            doc_type=constants.BIOME_METADATA_INDEX_DOC,
            id=self.es_index,
            body={
                "doc": dict(
                    name=self.es_index, created_at=datetime.datetime.now(), **parameters
                ),
                "doc_as_upsert": True,
            },
        )

    def create_explore_data_index(self, force_delete: bool):
        """Creates an explore data index if not exists or is forced"""
        dynamic_templates = [
            {
                data_type: {
                    "match_mapping_type": data_type,
                    "path_match": path_match,
                    "mapping": {
                        "type": "text",
                        "fields": {"keyword": {"type": "keyword", "ignore_above": 256}},
                    },
                }
            }
            for data_type, path_match in [("*", "*.value"), ("string", "*")]
        ]

        if force_delete:
            self.client.indices.delete(index=self.es_index, ignore=[400, 404])

        self.client.indices.create(
            index=self.es_index,
            body={"mappings": {self.es_doc: {"dynamic_templates": dynamic_templates}}},
            ignore=400,
        )


class TrainConfiguration:
    """Configures a training run

    Parameters
    ----------
        output: `str`
             The experiment output path
        trainer: `TrainerConfiguration`
             The trainer file path
        training: `str`
            The training datasource file path
        validation: `Optional[str]`
            The validation datasource file path
        test: `Optional[str]`
            The test datasource file path
    """

    def __init__(
        self,
        output: str,
        trainer: TrainerConfiguration,
        training: str = "",
        validation: Optional[str] = None,
        test: Optional[str] = None,
    ):
        self.output = output
        self.trainer = trainer
        self.training = training
        self.validation = validation
        self.test = test



from typing import Any, Dict, List, Optional, Union, cast

import numpy
import torch
from allennlp.data import Batch, Instance, TextFieldTensors
from allennlp.data.fields import TextField
from allennlp.modules.seq2vec_encoders import BagOfEmbeddingsEncoder
from allennlp.nn.util import get_text_field_mask
from captum.attr import IntegratedGradients

from biome.text import vocabulary
from biome.text.backbone import ModelBackbone
from biome.text.modules.specs import (
    ComponentSpec,
    FeedForwardSpec,
    Seq2VecEncoderSpec,
)
from .classification import ClassificationHead
from ..task_head import TaskOutput


class TextClassification(ClassificationHead):
    """
    Task head for text classification
    """

    label_name = "label"
    forward_arg_name = "text"

    def __init__(
        self,
        backbone: ModelBackbone,
        labels: List[str],
        pooler: Optional[Seq2VecEncoderSpec] = None,
        feedforward: Optional[FeedForwardSpec] = None,
        multilabel: bool = False,
    ) -> None:

        super(TextClassification, self).__init__(backbone, labels, multilabel)

        self.pooler = (
            pooler.input_dim(self.backbone.encoder.get_output_dim()).compile()
            if pooler
            else BagOfEmbeddingsEncoder(self.backbone.encoder.get_output_dim())
        )
        self.feedforward = (
            None
            if not feedforward
            else feedforward.input_dim(self.pooler.get_output_dim()).compile()
        )
        self._classification_layer = torch.nn.Linear(
            (self.feedforward or self.pooler).get_output_dim(), self.num_labels
        )

    def featurize(
        self, text: Any, label: Optional[Union[int, str, List[Union[int, str]]]] = None
    ) -> Optional[Instance]:
        instance = self.backbone.featurizer(
            text,
            to_field=self.forward_arg_name,
            aggregate=True,
            exclude_record_keys=True,
        )
        return self.add_label(instance, label, to_field=self.label_name)

    def forward(  # type: ignore
        self, text: TextFieldTensors, label: torch.IntTensor = None,
    ) -> TaskOutput:

        mask = get_text_field_mask(text)
        embedded_text = self.backbone.forward(text, mask)
        embedded_text = self.pooler(embedded_text, mask=mask)

        if self.feedforward is not None:
            embedded_text = self.feedforward(embedded_text)

        logits = self._classification_layer(embedded_text)
        return self.calculate_output(logits=logits, label=label)

    def explain_prediction(
        self, prediction: Dict[str, numpy.array], instance: Instance
    ) -> Dict[str, Any]:

        dataset = Batch([instance])
        input_tokens_ids = dataset.as_tensor_dict()
        ig = IntegratedGradients(self._explain_embeddings)

        num_wrapping_dims = 0

        text_tokens = [
            token.text
            for token in cast(TextField, instance.get(self.forward_arg_name)).tokens
        ]
        text_tensor = input_tokens_ids.get(self.forward_arg_name)
        mask = get_text_field_mask(text_tensor, num_wrapping_dims=num_wrapping_dims)
        text_embeddings = self.backbone.embedder.forward(
            text_tensor, num_wrapping_dims=num_wrapping_dims
        )

        label_id = vocabulary.index_for_label(
            self.backbone.vocab, prediction.get(self.label_name)
        )
        attributions, delta = ig.attribute(
            text_embeddings,
            target=label_id,
            additional_forward_args=mask,
            return_convergence_delta=True,
        )
        # TODO: what the attribution and deltas means
        attributions = attributions.sum(dim=2).squeeze(0)
        attributions = attributions / torch.norm(attributions)
        attributions = attributions.detach().numpy()

        return {
            **prediction,
            "explain": {
                self.forward_arg_name: [
                    {"token": token, "attribution": attribution}
                    for token, attribution in zip(text_tokens, attributions)
                ]
            },
        }

    def _explain_embeddings(
        self, embeddings: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """Explain embeddings for single text classification task"""
        embedded_text = self.backbone.encoder.forward(embeddings, mask)
        embedded_text = self.pooler.forward(embedded_text)
        if self.feedforward:
            embedded_text = self.feedforward.forward(embedded_text)
        logits = self._classification_layer(embedded_text)
        # TODO: review what kind of information we need to pass
        return logits


class TextClassificationSpec(ComponentSpec[TextClassification]):
    """Spec for classification head components"""

    pass



import json
import random
from pathlib import Path
from typing import Tuple

import numpy as np
import pytest
import torch
import yaml
from biome.text import Pipeline, TrainerConfiguration, VocabularyConfiguration
from biome.text.data import DataSource
from biome.text.configuration import WordFeatures
from biome.text.configuration import CharFeatures


@pytest.fixture
def train_valid_data_source() -> Tuple[DataSource, DataSource]:
    resources_path = Path(__file__).parent / "resources"
    training_ds = DataSource(source=str(resources_path / "business.cat.2k.train.csv"))
    validation_ds = DataSource(source=str(resources_path / "business.cat.2k.valid.csv"))

    return training_ds, validation_ds


@pytest.fixture
def pipeline_dict() -> dict:
    """Pipeline config dict. You need to update the labels!"""
    pipeline_dict = {
        "name": "german_business_names",
        "features": {
            "word": {"embedding_dim": 16, "lowercase_tokens": True,},
            "char": {
                "embedding_dim": 16,
                "encoder": {
                    "type": "gru",
                    "num_layers": 1,
                    "hidden_size": 32,
                    "bidirectional": True,
                },
                "dropout": 0.1,
            },
        },
        "head": {
            "type": "TextClassification",
            "labels": [
                "Unternehmensberatungen",
                "Friseure",
                "Tiefbau",
                "Dienstleistungen",
                "Gebrauchtwagen",
                "Restaurants",
                "Architekturbüros",
                "Elektriker",
                "Vereine",
                "Versicherungsvermittler",
                "Sanitärinstallationen",
                "Edv",
                "Maler",
                "Physiotherapie",
                "Werbeagenturen",
                "Apotheken",
                "Vermittlungen",
                "Hotels",
                "Autowerkstätten",
                "Elektrotechnik",
                "Allgemeinärzte",
                "Handelsvermittler Und -vertreter",
            ],
            "pooler": {
                "type": "gru",
                "num_layers": 1,
                "hidden_size": 16,
                "bidirectional": True,
            },
            "feedforward": {
                "num_layers": 1,
                "hidden_dims": [16],
                "activations": ["relu"],
                "dropout": [0.1],
            },
        },
    }

    return pipeline_dict


@pytest.fixture
def trainer_dict() -> dict:
    return {
        "batch_size": 64,
        "num_epochs": 5,
        "optimizer": {"type": "adam", "lr": 0.01,},
    }


def test_text_classification(
    tmp_path, pipeline_dict, trainer_dict, train_valid_data_source
):
    """Apart from a well specified training, this also tests the vocab creation!"""
    random.seed(42)
    np.random.seed(422)
    torch.manual_seed(4222)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(4222)

    pl = Pipeline.from_config(pipeline_dict)
    trainer = TrainerConfiguration(**trainer_dict)
    vocab = VocabularyConfiguration(
        sources=[train_valid_data_source[0]], max_vocab_size={"word": 50}
    )

    pl.create_vocabulary(vocab)

    assert pl._model.vocab.get_vocab_size(WordFeatures.namespace) == 52
    assert pl._model.vocab.get_vocab_size(CharFeatures.namespace) == 83

    output = tmp_path / "output"

    pl.train(
        output=str(output),
        trainer=trainer,
        training=train_valid_data_source[0],
        validation=train_valid_data_source[1],
    )

    assert pl.trainable_parameters == 22070

    with (output / "metrics.json").open() as file:
        metrics = json.load(file)

    assert metrics["training_loss"] == pytest.approx(0.767, abs=0.003)

    # test vocab from a pretrained file
    pl = Pipeline.from_pretrained(str(output / "model.tar.gz"))

    assert pl._model.vocab.get_vocab_size(WordFeatures.namespace) == 52
    assert pl._model.vocab.get_vocab_size(CharFeatures.namespace) == 83



import os

from .environment import BIOME_EXPLORE_ENDPOINT, ES_HOST

BIOME_METADATA_INDEX = ".biome"
BIOME_METADATA_INDEX_DOC = "_doc"
# This is the biome explore UI endpoint, used for show information
# about explorations once the data is persisted
EXPLORE_APP_ENDPOINT = os.getenv(BIOME_EXPLORE_ENDPOINT, "http://localhost:8080")

DEFAULT_ES_HOST = os.getenv(ES_HOST, "localhost:9200")
